@@ -1561,6 +1561,17 @@ class TestSync():
             with open(outfile) as f:
                 assert json.loads(f.read()) == {"paths": ["/index_html/"]}
 
+    def addscript(self, basename, *lines):
+        """
+        Add executable file into zeo dir, returning the full filename
+        """
+        fname = "{}/{}".format(self.zeo.path, basename)
+        lines = ("#!/bin/bash",) + lines
+        with open(fname, 'w') as f:
+            f.write('\n'.join(lines))
+        os.chmod(fname, 0o700)
+        return fname
+
     def test_playback_hook(self):
         """
         Add configuration option for a playback hook script and check that
@@ -1572,44 +1583,25 @@ class TestSync():
         # Reset the commit
         self.gitrun('reset', '--hard', 'HEAD~2')
 
-        playback_cmd = "{}/playback_cmd".format(self.zeo.path)
-        cmd_script = '\n'.join([
-            "#!/bin/bash",
-            "cat > {}"
-        ]).format('{}.out'.format(playback_cmd))
-        with open(playback_cmd, 'w') as f:
-            f.write(cmd_script)
-        os.chmod(playback_cmd, 0o700)
+        playback_cmd_out = "{}/playback_cmd.out".format(self.zeo.path)
+        playback_cmd = self.addscript(
+            "playback_cmd",
+            "cat > {}".format(playback_cmd_out),
+        )
 
-        fname = "{}/playback_hook".format(self.zeo.path)
-        playback_dict = [{
-            "paths": ["/NewFolder"],
-            "cmd": playback_cmd
-        }]
-
-        script = '\n'.join([
-            "#!/bin/bash",
-            "echo '{}'".format(json.dumps(playback_dict)),
-        ])
-        with open(fname, 'w') as f:
-            f.write(script)
-        os.chmod(fname, 0o700)
-        with open(self.config.path) as f:
-            orig_config = f.read()
-        with open(self.config.path, 'a') as f:
-            f.write('\nplayback_hook = "{}"\n'.format(fname))
-
-        # Avoid error regarding reusing runner with changed config
-        del self.runner
-        self.run('pick', 'HEAD..{}'.format(commit))
+        playback_hook = self.addscript(
+            "playback_hook",
+            "echo '{}'".format(json.dumps([{
+                "paths": ["/NewFolder"],
+                "cmd": playback_cmd,
+            }])),
+        )
+        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
+            self.run('pick', 'HEAD..{}'.format(commit))
 
         assert 'NewFolder' in self.app.objectIds()
         assert 'NewFolder2' not in self.app.objectIds()
-        assert os.path.isfile('{}.out'.format(playback_cmd))
-
-        with open(self.config.path, 'w') as f:
-            f.write(orig_config)
-        del self.runner
+        assert os.path.isfile(playback_cmd_out)
 
     def test_playback_hook_failed(self):
         """
@@ -1622,31 +1614,23 @@ class TestSync():
         # Reset the commit
         self.gitrun('reset', '--hard', 'HEAD~2')
 
-        playback_cmd = "{}/playback_cmd".format(self.zeo.path)
-        cmd_script = '\n'.join([
-            "#!/bin/bash",
-            "exit 42"
-        ])
-        with open(playback_cmd, 'w') as f:
-            f.write(cmd_script)
-        os.chmod(playback_cmd, 0o700)
-
-        fname = "{}/playback_hook".format(self.zeo.path)
-        playback_dict = [{
-            "paths": ["/NewFolder"],
-            "cmd": playback_cmd
-            }, {
-            "paths": ["/NewFolder2"],
-            },
-        ]
-        script = '\n'.join([
-            "#!/bin/bash",
-            "echo '{}'".format(json.dumps(playback_dict)),
-        ])
-        with open(fname, 'w') as f:
-            f.write(script)
-        os.chmod(fname, 0o700)
-        with self.appendtoconf('playback_hook = "{}"'.format(fname)):
+        playback_cmd = self.addscript(
+            "playback_cmd",
+            "exit 42",
+        )
+        playback_hook = self.addscript(
+            "playback_hook",
+            "echo '{}'".format(json.dumps([
+                {
+                    "paths": ["/NewFolder"],
+                    "cmd": playback_cmd,
+                },
+                {
+                    "paths": ["/NewFolder2"],
+                },
+            ])),
+        )
+        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
             with pytest.raises(AssertionError):
                 self.run('pick', 'HEAD..{}'.format(commit))
 
@@ -1654,7 +1638,7 @@ class TestSync():
             assert 'NewFolder2' not in self.app.objectIds()
 
     @contextmanager
-    def addlayer(self, seqnum='00'):
+    def addlayer(self, seqnum='00', tarsource=False):
         """
         Create a temp directory and add a config that uses this as additional
         code layer.
@@ -1664,9 +1648,23 @@ class TestSync():
         ))
         path = '{}/layers/{}'.format(self.config.folder, name)
         with tempfile.TemporaryDirectory() as layer:
+            workdir = f'{layer}/workdir'
+            os.makedirs(f'{workdir}/__root__')
+            subprocess.run(['git', 'init'], cwd=workdir)
+            subprocess.run(['git', 'config', 'user.email',
+                            'zodbsync-tester@perfact.de'], cwd=workdir)
+            subprocess.run(['git', 'config', 'user.name',
+                            'ZODBSync tester'], cwd=workdir)
+            os.makedirs(f'{layer}/source/__root__')
+            with open(f'{layer}/source.tar', 'w'):
+                pass
+            if tarsource:
+                source = f'{layer}/source.tar'
+            else:
+                source = f'{layer}/source'
             with open(path, 'w') as f:
-                f.write('base_dir = "{}"\n'.format(layer))
-            os.mkdir(os.path.join(layer, '__root__'))
+                f.write(f'workdir = "{layer}/workdir"\n')
+                f.write(f'source = "{source}"\n')
             # Force re-reading config
             if hasattr(self, 'runner'):
                 del self.runner
@@ -1688,7 +1686,7 @@ class TestSync():
         with self.addlayer() as layer:
             shutil.copytree(
                 '{}/__root__/Test'.format(self.repo.path),
-                '{}/__root__/Test'.format(layer),
+                '{}/workdir/__root__/Test'.format(layer),
             )
             self.run('freeze', '/')
             self.run('record', '/')
@@ -1707,7 +1705,7 @@ class TestSync():
         with self.addlayer() as layer:
             shutil.copytree(
                 '{}/__root__/Test'.format(self.repo.path),
-                '{}/__root__/Test'.format(layer),
+                '{}/workdir/__root__/Test'.format(layer),
             )
             self.run('record', '/')
         assert not os.path.exists(
@@ -1732,14 +1730,14 @@ class TestSync():
         with self.addlayer() as layer:
             shutil.copytree(
                 '{}/__root__/Test'.format(self.repo.path),  # custom layer!
-                '{}/__root__/Test'.format(layer),           # new base layer!
+                '{}/workdir/__root__/Test'.format(layer),  # new base layer!
             )
             # now create the standard Test folder titled 'Something
             meta = zodbsync.mod_format({
                 'title': 'Something',
                 'type': 'Folder'
             })
-            with open(os.path.join(layer, '__root__/Test/__meta__'), 'w') as f:
+            with open(f'{layer}/workdir/__root__/Test/__meta__', 'w') as f:
                 f.write(meta)
             self.run('playback', '/')
 
@@ -1764,7 +1762,7 @@ class TestSync():
         self.add_folder('Test')
         with self.addlayer() as layer:
             src = '{}/__root__'.format(self.repo.path)
-            tgt = '{}/__root__'.format(layer)
+            tgt = '{}/workdir/__root__'.format(layer)
             os.rename(src + '/Test', tgt + '/Test')
             cmd = ['playback', '/Test']
             if not recurse:
@@ -1780,7 +1778,7 @@ class TestSync():
         self.add_folder('Test')
         with self.addlayer() as layer:
             src = '{}/__root__'.format(self.repo.path)
-            tgt = '{}/__root__'.format(layer)
+            tgt = '{}/workdir/__root__'.format(layer)
             shutil.copytree(src + '/Test', tgt + '/Test')
             with open('{}/__frozen__'.format(src), 'w'):
                 pass
@@ -1802,7 +1800,7 @@ class TestSync():
         with self.addlayer() as layer:
             root = os.path.join(self.repo.path, '__root__')
             # Move current structure into lower layer
-            os.rename(root, os.path.join(layer, '__root__'))
+            os.rename(root, os.path.join(layer, 'workdir/__root__'))
             # Create a sparse structure in top layer
             files = [
                 'Test1/__frozen__',
@@ -1842,7 +1840,7 @@ class TestSync():
         self.run('record', '/Test')
         with self.addlayer() as layer:
             root = [
-                os.path.join(layer, '__root__'),
+                os.path.join(layer, 'workdir/__root__'),
                 os.path.join(self.repo.path, '__root__'),
             ]
             os.rename(os.path.join(root[1], 'Test'),
@@ -1860,7 +1858,7 @@ class TestSync():
         self.add_folder('Sub', parent='Test')
         with self.addlayer() as layer:
             srcroot = os.path.join(self.repo.path, '__root__')
-            tgtroot = os.path.join(layer, '__root__')
+            tgtroot = os.path.join(layer, 'workdir/__root__')
             os.rename(os.path.join(srcroot, 'Test'),
                       os.path.join(tgtroot, 'Test'))
             self.run('record', '/')
@@ -1878,7 +1876,7 @@ class TestSync():
         with self.addlayer() as layer:
             os.rename(
                 os.path.join(self.repo.path, '__root__/__meta__'),
-                os.path.join(layer, '__root__/__meta__'),
+                os.path.join(layer, 'workdir/__root__/__meta__'),
             )
             self.run('record', '/')
         assert not os.path.isdir(
@@ -1894,7 +1892,7 @@ class TestSync():
         with self.addlayer() as layer:
             os.rename(
                 os.path.join(self.repo.path, '__root__/index_html'),
-                os.path.join(layer, '__root__/index_html'),
+                os.path.join(layer, 'workdir/__root__/index_html'),
             )
             watcher = self.mkrunner('watch')
             watcher.setup()
@@ -1927,7 +1925,7 @@ class TestSync():
         self.run('record', '/')
         with self.addlayer() as layer:
             src = os.path.join(self.repo.path, '__root__')
-            tgt = os.path.join(layer, '__root__')
+            tgt = os.path.join(layer, 'workdir/__root__')
             os.rmdir(tgt)
             os.rename(src, tgt)
             os.mkdir(src)
@@ -1967,7 +1965,7 @@ class TestSync():
             root = os.path.join(self.repo.path, '__root__')
             os.rename(
                 os.path.join(root, 'Test'),
-                os.path.join(layer, '__root__/Test'),
+                os.path.join(layer, 'workdir/__root__/Test'),
             )
             self.app.manage_delObjects(ids=['Test'])
             self.run('record', '/')
@@ -1992,7 +1990,7 @@ class TestSync():
             root = os.path.join(self.repo.path, '__root__')
             os.rename(
                 os.path.join(root, 'Test'),
-                os.path.join(layer, '__root__/Test'),
+                os.path.join(layer, 'workdir/__root__/Test'),
             )
             self.app.Test.manage_delObjects(ids=['Sub'])
             self.run('record', '/')
@@ -2012,18 +2010,16 @@ class TestSync():
             self.run('record', '/')
             ident = self.runner.sync.layers[-1]['ident']
             src = os.path.join(self.repo.path, '__root__')
-            tgt = os.path.join(layer, '__root__')
+            tgt = os.path.join(layer, 'source/__root__')
             os.rmdir(tgt)
             os.rename(src, tgt)
             os.mkdir(src)
-            self.run('layer-hash', layer)
-            self.run('layer-init')
+            self.run('layer-init', '*')
             with open(os.path.join(tgt, 'Test/__meta__'), 'w') as f:
                 f.write(zodbsync.mod_format({
                     'title': 'Changed',
                     'type': 'Folder'
                 }))
-            self.run('layer-hash', layer)
             self.run('layer-update', ident)
             assert 'Conflict with object' not in caplog.text
             assert self.app.Test.title == 'Changed'
@@ -2129,12 +2125,11 @@ class TestSync():
             self.run('record', '/')
             ident = self.runner.sync.layers[-1]['ident']
             src = os.path.join(self.repo.path, '__root__')
-            tgt = os.path.join(layer, '__root__')
+            tgt = os.path.join(layer, 'source/__root__')
             os.rmdir(tgt)
             os.rename(src, tgt)
             os.mkdir(src)
-            self.run('layer-hash', layer)
-            self.run('layer-init')
+            self.run('layer-init', '*')
             with self.runner.sync.tm:
                 self.app.Test._setProperty('nav_hidden', True, 'boolean')
                 self.app.ToDelete.Sub._setProperty('nav_hidden', True,
@@ -2146,9 +2141,143 @@ class TestSync():
                     'type': 'Folder'
                 }))
             shutil.rmtree(os.path.join(tgt, 'ToDelete'))
-            self.run('layer-hash', layer)
             self.run('layer-update', ident)
             expect = 'Conflict with object in custom layer: '
             assert expect + '/Test' in caplog.text
             assert 'AttributeError' not in caplog.text
             assert expect + '/ToDelete/Sub' in caplog.text
+
+    def test_layer_change_into_top(self):
+        """
+        Verify that changed files are written into the top layer.
+        Note that this is not what we want in the long run, but until we have
+        methods for moving objects between layers and there is a frontend for
+        showing unstaged changes in all layers, everything is written into the
+        top layer.
+        """
+        with self.runner.sync.tm:
+            self.app.manage_addProduct['OFSP'].manage_addFile(id='blob')
+
+        with self.addlayer() as layer:
+            self.run('record', '/blob')
+            shutil.move(
+                '{}/__root__/blob'.format(self.repo.path),
+                '{}/workdir/__root__/blob'.format(layer),
+            )
+            with self.runner.sync.tm:
+                self.app.blob.manage_edit(
+                    filedata='text_content',
+                    content_type='text/plain',
+                    title='BLOB'
+                )
+            self.run('record', '/')
+            root = os.path.join(self.repo.path, '__root__')
+            # both meta and source file are in custom layer
+            assert os.path.exists(os.path.join(root, 'blob/__meta__'))
+            assert os.path.exists(os.path.join(root, 'blob/__source__.txt'))
+            source_fmt = '{}/__root__/blob/__source__.txt'
+            with open(source_fmt.format(f'{layer}/workdir')) as f:
+                # source in layer should still be empty
+                assert f.read() == ''
+            with open(source_fmt.format(self.repo.path)) as f:
+                # ... content is in custom layer!
+                assert f.read() == 'text_content'
+
+    def test_layer_playback_hook(self):
+        """
+        Set up two layers. Pick a commit that marks an object as __deleted__ in
+        the top layer. Check that the playback hook script gets the normalized
+        object paths and not the specific files.
+        """
+        with self.runner.sync.tm:
+            self.app.manage_addProduct['OFSP'].manage_addFile(id='blob')
+
+        root = '{}/__root__'.format(self.repo.path)
+        with self.addlayer() as layer:
+            self.run('record', '/blob')
+            shutil.move(
+                '{}/blob'.format(root),
+                '{}/workdir/__root__/blob'.format(layer),
+            )
+            os.mkdir('{}/blob'.format(root))
+            with open('{}/blob/__deleted__'.format(root), 'w'):
+                pass
+            self.gitrun('add', '.')
+            self.gitrun('commit', '-m', 'delete blob')
+            commid = self.get_head_id()
+            self.gitrun('reset', '--hard', 'HEAD~')
+            output = '{}/playback_hook.out'.format(self.zeo.path)
+            playback_hook = self.addscript(
+                "playback_hook",
+                "cat > {}".format(output),
+                "echo '[]'",
+            )
+            with self.appendtoconf(
+                    'playback_hook = "{}"'.format(playback_hook)
+            ):
+                self.run('pick', commid)
+            with open(output) as f:
+                assert {"paths": ["/blob/"]} == json.loads(f.read())
+
+    def test_layer_tar(self):
+        """
+        Perform a layer-init and layer-update from a tar file source.
+        """
+        with self.runner.sync.tm:
+            self.app.manage_addProduct['OFSP'].manage_addFile(id='blob')
+        with self.addlayer(tarsource=True) as layer:
+            self.run('record', '/blob')
+            subprocess.run(
+                ['tar', 'cf', f'{layer}/source.tar', 'blob'],
+                cwd=f'{self.repo.path}/__root__',
+                check=True,
+            )
+            self.run('layer-init', '*')
+            assert os.listdir(f'{layer}/workdir/__root__') == ['blob']
+            # Record to remove from fallback layer
+            self.run('record', '/')
+            assert 'blob' not in os.listdir(f'{self.repo.path}/__root__')
+            # Now change the file in the TAR file and run layer-update
+            shutil.copytree(
+                f'{layer}/workdir/__root__/blob',
+                f'{layer}/blob',
+            )
+            with open(f'{layer}/blob/__source__.txt', 'w') as f:
+                f.write('changed')
+            subprocess.run(
+                ['tar', 'cf', f'{layer}/source.tar', 'blob'],
+                cwd=layer,
+                check=True,
+            )
+            self.run('layer-update', '*')
+            assert str(self.app.blob) == 'changed'
+
+    def test_layer_update_2phase_failed(self):
+        """
+        Perform layer-update with a two-phase playback where a command at the
+        end of the first phase fails. Check that the rollback is performed
+        correctly.
+        """
+        with self.runner.sync.tm:
+            self.app.manage_addProduct['OFSP'].manage_addFile(id='blob')
+        playback_cmd = self.addscript("playback_cmd", "false")
+
+        playback_hook = self.addscript(
+            "playback_hook",
+            "echo '{}'".format(json.dumps([{
+                "paths": ["/blob"],
+                "cmd": playback_cmd,
+            }])),
+        )
+        with self.appendtoconf('playback_hook = "{}"'.format(playback_hook)):
+            with self.addlayer() as layer:
+                self.run('record', '/')
+                src = f'{self.repo.path}/__root__/blob'
+                tgt = f'{layer}/source/__root__/blob'
+                os.rename(src, tgt)
+                self.run('layer-init', '*')
+                with open(f'{tgt}/__source__.txt', 'w') as f:
+                    f.write('changed')
+                with pytest.raises(AssertionError):
+                    self.run('layer-update', '*')
+                assert str(self.app.blob) == ''
